/****************************************************************************
 *                                                                           *
 * xsublim -- Submit.  Conform.  Obey.                                       *
 *                                                                           *
 * Copyright (c) 1999 - 2000 Greg Knauss (greg@eod.com)                      *
 *                                                                           *
 * Thanks to Jamie Zawinski, whose suggestions and advice made what was a    *
 * boring little program into a less boring (and a _lot_ less little)        *
 * program.                                                                  *
 *                                                                           *
 * Permission to use, copy, modify, distribute, and sell this software and   *
 * its documentation for any purpose is hereby granted without fee, provided *
 * that the above copyright notice appear in all copies and that both that   *
 * copyright notice and this permission notice appear in supporting          *
 * documentation.  No representations are made about the suitability of this *
 * software for any purpose.  It is provided "as is" without express or      *
 * implied warranty.                                                         *
 *                                                                           *
 * Stare into the subliminal for as long as you can...                       *
 *  																		 *
 * ***************************************************************************
 * Copyright (c)2010 by Markus Schulz (schulz@alpharesearch.de)              *
 * All my changes are under GNU General Public License.                      *
 *****************************************************************************/

/* Defines *******************************************************************/
#define XSUBLIM_NAME               "XSublim"
#define XSUBLIM_TEXT_COUNT         1000
#define XSUBLIM_TEXT_LENGTH        128
#define XSUBLIM_TEXT_OUTLINE       1
#define XSUBLIM_PROGRAM_SIZE       1024*10
#define XSUBLIM_ARG_DELAYSHOW      "delayShow"
#define XSUBLIM_ARG_DELAYWORD      "delayWord"
#define XSUBLIM_ARG_DELAYPHRASEMIN "delayPhraseMin"
#define XSUBLIM_ARG_DELAYPHRASEMAX "delayPhraseMax"
#define XSUBLIM_ARG_RANDOM         "random"
#define XSUBLIM_ARG_FILE           "file"
#define XSUBLIM_ARG_PROGRAM        "program"
#define XSUBLIM_ARG_SCREENSAVER    "screensaver"
#define XSUBLIM_ARG_OUTLINE        "outline"
#define XSUBLIM_ARG_CENTER         "center"
#define XSUBLIM_ARG_FONT           "font"
#define XSUBLIM_ARG_PHRASES        "phrases"
#define XSUBLIM_ARG_HELP           "help"
#define XSUBLIM_ARG_WORDMODE       "wordmode"
#define XSUBLIM_ARG_OFFSET_X       "offsetx"
#define XSUBLIM_ARG_OFFSET_Y       "offsety"
#define XSUBLIM_ARG_ONESHOT        "oneshot"
#ifndef TRUE
#define FALSE 0
#define TRUE  1
#endif


/* Includes ******************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/stat.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/CoreP.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xos.h>
#include <X11/Xproto.h>
#if defined(__sgi)
#include <X11/SGIScheme.h>
#endif

#include "usleep.h"
#include "yarandom.h"
#include "resources.h"
#include "vroot.h"
#include "../config.h"


/* Globals *******************************************************************/
char *progname;
char *progclass = XSUBLIM_NAME;

# ifdef __GNUC__
__extension__                   /* don't warn about "string length is greater than the
                                   length ISO C89 compilers are required to support"
                                   in the following string constant... */
# endif
static char *defaults[] = {
  ".background:                     #000000",
  ".foreground:                     #FFFFFF",
  "*" XSUBLIM_ARG_PHRASES ":"
    "I am healthy and happy.\\n"
    "Wealth is pouring into my life.\\n"
    "I am sailing on the river of wealth.\\n"
    "I am getting wealthier each day.\\n"
    "My body is healthy and functioning in a very good way.\\n"
    "I have a lot of energy.\\n"
    "I study and comprehend fast.\\n"
    "My mind is calm.\\n"
    "I am calm and relaxed in every situation.\\n"
    "My thoughts are under my control.\\n"
    "I radiate love and happiness.\\n"
    "I am surrounded by love.\\n"
    "I have the perfect job for me.\\n"
    "I am living in the house of my dreams.\\n"
    "I have good and loving relations with my spouse.\\n"
    "I have a wonderful and satisfying job.\\n"
    "I have the means to travel abroad, whenever I want to.\\n"
    "I am successful in whatever I do.\\n"
    "Everything is getting better every day.\\n"
    "All the things I want and need come to me.\\n"
    "I always receive more than what I need.\\n"
    "I have a bank account with more than enough.\\n"
    "I am an abundant person.\\n"
    "I create abundance in all that I say and do.\\n"
    "I accept abundance.\\n"
    "I welcome, and am open to receive all abundance that comes.\\n"
    "I draw abundance to myself today and every day.\\n"
    "I am successful. \\n"
    "Everything I do turns into success.\\n"
    "I am filled with success.\\n"
    "Success comes effortlessly to my direction.\\n"
    "My success is contagious, other people like it, seek it and respect it.\\n"
    "I attract positive-minded people to me.\\n"
    "I draw all things positive to myself.\\n"
    "I am very fortunate to work at what I love to do.\\n"
    "I am glowing with health and wholeness.\\n"
    "I behave in ways that promote my health more every day.\\n"
    "I deserve to be in perfect health.\\n"
    "I am healthy since my practices are healthy.\\n"
    "I love nutritious healthy food.\\n" "I enjoy eating fresh fruits and vegetables.\\n"
    "I am caring, smart, supportive, loyal, and fun to be with.\\n",
  "*" XSUBLIM_ARG_FONT ":           -*-utopia-*-r-*-*-*-600-*-*-p-*-*-*",
  "*" XSUBLIM_ARG_DELAYSHOW ":      40000",
  "*" XSUBLIM_ARG_DELAYWORD ":      100000",
  "*" XSUBLIM_ARG_DELAYPHRASEMIN ": 5000000",
  "*" XSUBLIM_ARG_DELAYPHRASEMAX ": 20000000",
  "*" XSUBLIM_ARG_OFFSET_X ":       1.0",
  "*" XSUBLIM_ARG_OFFSET_Y":        1.0",
  "*" XSUBLIM_ARG_RANDOM ":         true",
  "*" XSUBLIM_ARG_SCREENSAVER ":    false",
  "*" XSUBLIM_ARG_OUTLINE ":        true",
  "*" XSUBLIM_ARG_CENTER ":         true",
  "*" XSUBLIM_ARG_HELP ":           false",
  "*" XSUBLIM_ARG_WORDMODE ":       false",
  "*" XSUBLIM_ARG_ONESHOT":         false",
  NULL
};

static XrmOptionDescRec options[] = {
  {"--" XSUBLIM_ARG_HELP, "." XSUBLIM_ARG_HELP, XrmoptionNoArg, "true"},
  {"--" XSUBLIM_ARG_FONT, "." XSUBLIM_ARG_FONT, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_DELAYSHOW, "." XSUBLIM_ARG_DELAYSHOW, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_DELAYWORD, "." XSUBLIM_ARG_DELAYWORD, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_DELAYPHRASEMIN, "." XSUBLIM_ARG_DELAYPHRASEMIN, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_DELAYPHRASEMAX, "." XSUBLIM_ARG_DELAYPHRASEMAX, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_OFFSET_X, "." XSUBLIM_ARG_OFFSET_X, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_OFFSET_Y, "." XSUBLIM_ARG_OFFSET_Y, XrmoptionSepArg, 0},
  {"--" XSUBLIM_ARG_RANDOM, "." XSUBLIM_ARG_RANDOM, XrmoptionNoArg, "true"},
  {"--no-" XSUBLIM_ARG_RANDOM, "." XSUBLIM_ARG_RANDOM, XrmoptionNoArg, "false"},
  {"--" XSUBLIM_ARG_FILE, "." XSUBLIM_ARG_FILE, XrmoptionSepArg, 0},
#if !defined(VMS)
  {"--" XSUBLIM_ARG_PROGRAM, "." XSUBLIM_ARG_PROGRAM, XrmoptionSepArg, 0},
#endif
  {"--" XSUBLIM_ARG_SCREENSAVER, "." XSUBLIM_ARG_SCREENSAVER, XrmoptionNoArg, "true"},
  {"--no-" XSUBLIM_ARG_SCREENSAVER, "." XSUBLIM_ARG_SCREENSAVER, XrmoptionNoArg, "false"},
  {"--" XSUBLIM_ARG_OUTLINE, "." XSUBLIM_ARG_OUTLINE, XrmoptionNoArg, "true"},
  {"--no-" XSUBLIM_ARG_OUTLINE, "." XSUBLIM_ARG_OUTLINE, XrmoptionNoArg, "false"},
  {"--" XSUBLIM_ARG_CENTER, "." XSUBLIM_ARG_CENTER, XrmoptionNoArg, "true"},
  {"--no-" XSUBLIM_ARG_CENTER, "." XSUBLIM_ARG_CENTER, XrmoptionNoArg, "false"},
  {"--" XSUBLIM_ARG_WORDMODE, "." XSUBLIM_ARG_WORDMODE, XrmoptionNoArg, "true"},
  {"--" XSUBLIM_ARG_ONESHOT, "." XSUBLIM_ARG_ONESHOT, XrmoptionNoArg, "true"},
  {NULL, NULL, 0, 0}
};

static int Xsublim_Sig_Last;


/* Functions *****************************************************************/

/* Defer signals to protect the server grab ================================ */
static void
xsublim_Sig_Catch (int sig_Number)
{
  /* BSD needs this reset each time, and it shouldn't hurt anything
     else */
  signal (sig_Number, xsublim_Sig_Catch);
  Xsublim_Sig_Last = sig_Number;
}

/* This was all basically swiped from driver/remote.c and util/vroot.h */
static Window
xsublim_Ss_GetWindow (Display * ss_Display)
{
  Screen *s = DefaultScreenOfDisplay (ss_Display);
  Window root = XRootWindowOfScreen (s);
  Window vroot = VirtualRootWindowOfScreen (s);
  if (root == vroot)
    return 0;
  else
    return vroot;
}

/* Main ==================================================================== */
static XErrorHandler Xsublim_Sh_Handler = NULL;
static int Xsublim_Sh_Status = 0;

static int
xsublim_Sh_Handler (Display * handle_Display, XErrorEvent * handle_Error)
{
  if (handle_Error->error_code == BadMatch)
    {
      Xsublim_Sh_Status = BadMatch;
      return 0;
    }
  if (Xsublim_Sh_Handler == NULL)
    {
      fprintf (stderr, "%s: ", progname);
      abort ();
    }
  return (*Xsublim_Sh_Handler) (handle_Display, handle_Error);
}


int
main (int argc, char *argv[])
{
  XtAppContext app;
  int sig_Number;
  int sig_Signal[] = {
    SIGHUP,
    SIGINT,
    SIGQUIT,
    SIGILL,
    SIGTRAP,
#if defined(SIGIOT)
    SIGIOT,
#endif
    SIGABRT,
#if defined(SIGEMT)
    SIGEMT,
#endif
    SIGFPE,
    SIGBUS,
    SIGSEGV,
#if defined(SIGSYS)
    SIGSYS,
#endif
    SIGTERM,
#if defined(SIGXCPU)
    SIGXCPU,
#endif
#if defined(SIGXFSZ)
    SIGXFSZ,
#endif
#if defined(SIGDANGER)
    SIGDANGER,
#endif
    -1
  };
  Widget app_App;
  Display *dpy;
  Window win_Root;
  XWindowAttributes attr_Win;
  XGCValues gc_ValFore;
  XGCValues gc_ValBack;
  GC gc_GcFore;
  GC gc_GcBack;
  XFontStruct *font_Font;
  char *font_List[] = {
    "-*-character-*-r-*-*-*-600-*-*-p-*-*-*",
    "-*-helvetica-*-r-*-*-*-600-*-*-p-*-*-*",
    "-*-lucida-*-r-*-*-*-600-*-*-p-*-*-*",
    "-*-*-*-r-*-sans-*-600-*-*-p-*-*-*",
    "-*-*-*-r-*-*-*-600-*-*-m-*-*-*",
    "-*-helvetica-*-r-*-*-*-240-*-*-p-*-*-*",
    "-*-lucida-*-r-*-*-*-240-*-*-p-*-*-*",
    "-*-*-*-r-*-sans-*-240-*-*-p-*-*-*",
    "-*-*-*-r-*-*-*-240-*-*-m-*-*-*",
    "fixed",
    NULL
  };
  int font_Index;
  int text_Length;
  int text_X;
  int text_Y;
  int text_Width;
  int text_Height;
  char *text_List[XSUBLIM_TEXT_COUNT];
  int text_Used[XSUBLIM_TEXT_COUNT];
  char text_Text[XSUBLIM_TEXT_LENGTH + 1];
  char *text_Phrase;
  char *text_Word;
  char *text_Wordmode;
  int text_Index;
  int text_Item;
  int text_Count;
  struct
  {
    int outline_X;
    int outline_Y;
  } text_Outline[] =
  {
    {
    -1, -1},
    {
    1, -1},
    {
    -1, 1},
    {
    1, 1},
    {
    0, 0}
  };
  int text_OutlineIndex;
  XImage *image_Image = NULL;
  int image_X = 0;
  int image_Y = 0;
  int image_Width = 0;
  int image_Height = 0;
  int arg_Count;
  int arg_FlagCenter;
  int arg_FlagHelp;
  int arg_FlagOutline;
  int arg_FlagScreensaver;
  int arg_FlagRandom;
  int arg_FlagWordmode;
  int arg_FlagOneshot;
  int arg_DelayShow;
  int arg_DelayWord;
  int arg_DelayPhraseMin;
  int arg_DelayPhraseMax;
  float arg_OffsetX;
  float arg_OffsetY;
  char *arg_Text;
  char *arg_Source;

  /* Set-up ---------------------------------------------------------- */

  /* Catch signals */
  Xsublim_Sig_Last = -1;
  for (sig_Number = 0; sig_Signal[sig_Number] != -1; sig_Number++)
    {
      signal (sig_Number, xsublim_Sig_Catch);
    }

  /* Randomize -- only need to do this here because this program
     doesn't use the `screenhack.h' or `lockmore.h' APIs. */
# undef ya_rand_init
  ya_rand_init (0);

  /* Handle all the X nonsense */
#if defined(__sgi)
  SgiUseSchemes ("none");
#endif
  for (arg_Count = 0; options[arg_Count].option != NULL; arg_Count++)
    {
      ;
    }
  app_App = XtAppInitialize (&app, progclass, options, arg_Count, &argc, argv, defaults, 0, 0);

  /* jwz */
  if (argc > 1)
    {
      int x = 18;
      int end = 78;
      int i;
      int count = (sizeof (options) / sizeof (*options)) - 1;
      fprintf (stderr, "Unrecognised option: %s\n", argv[1]);
      fprintf (stderr, "Options include: ");
      for (i = 0; i < count; i++)
        {
          char *sw = options[i].option;
          Bool argp = (options[i].argKind == XrmoptionSepArg);
          int size = strlen (sw) + (argp ? 6 : 0) + 2;
          if (x + size >= end)
            {
              fprintf (stderr, "\n\t\t ");
              x = 18;
            }
          x += size;
          fprintf (stderr, "%s", sw);
          if (argp)
            fprintf (stderr, " <arg>");
          if (i != count - 1)
            fprintf (stderr, ", ");
        }
      fprintf (stderr, ".\n");
      exit (-1);
    }

  dpy = XtDisplay (app_App);
  XtGetApplicationNameAndClass (dpy, &progname, &progclass);
  win_Root = RootWindowOfScreen (XtScreen (app_App));
  XtDestroyWidget (app_App);

  /* Get the arguments */
  arg_FlagCenter = get_boolean_resource (dpy, XSUBLIM_ARG_CENTER, "Boolean");
  arg_FlagHelp = get_boolean_resource (dpy, XSUBLIM_ARG_HELP, "Boolean");
  arg_FlagWordmode = get_boolean_resource (dpy, XSUBLIM_ARG_WORDMODE, "Boolean");
  arg_FlagOutline = get_boolean_resource (dpy, XSUBLIM_ARG_OUTLINE, "Boolean");
  arg_FlagScreensaver = get_boolean_resource (dpy, XSUBLIM_ARG_SCREENSAVER, "Boolean");
  arg_FlagRandom = get_boolean_resource (dpy, XSUBLIM_ARG_RANDOM, "Boolean");
  arg_DelayShow = get_integer_resource (dpy, XSUBLIM_ARG_DELAYSHOW, "Integer");
  arg_DelayWord = get_integer_resource (dpy, XSUBLIM_ARG_DELAYWORD, "Integer");
  arg_DelayPhraseMin = get_integer_resource (dpy, XSUBLIM_ARG_DELAYPHRASEMIN, "Integer");
  arg_DelayPhraseMax = get_integer_resource (dpy, XSUBLIM_ARG_DELAYPHRASEMAX, "Integer");
  arg_OffsetX = get_float_resource (dpy, XSUBLIM_ARG_OFFSET_X, "Float");
  arg_OffsetY = get_float_resource (dpy, XSUBLIM_ARG_OFFSET_Y, "Float");
  arg_FlagOneshot = get_boolean_resource (dpy, XSUBLIM_ARG_ONESHOT, "Boolean");
  
  if (arg_FlagHelp == 1)
    {
      fprintf (stdout, PACKAGE_STRING);
      fprintf (stdout, "\nUsage:\n");
      fprintf (stdout, "  xsublim [OPTION...]\n\n");
      fprintf (stdout, "Help Options:\n");
      fprintf (stdout, "  --help                  Show summary of options\n\n");
      fprintf (stdout, "Application Options:\n");
      fprintf (stdout, "  --font <font>           Font to use, default is utopia, character, helvetica, lucida, sans fixed\n");
      fprintf (stdout, "  --file <filename>       New-line delimited phrase file\n");
      fprintf (stdout, "  --program <executable>  New-line delimited phrase-producing executable\n");
      fprintf (stdout, "  --delayShow <ms>        Microsecs for display of each word, default: %i\n",arg_DelayShow);
      fprintf (stdout, "  --delayWord <ms>        Microsecs for blank between words, default: %i\n",arg_DelayWord);
      fprintf (stdout, "  --delayPhraseMin <ms>   Microsecs for min blank between phrases, default: %i\n",arg_DelayPhraseMin);
      fprintf (stdout, "  --delayPhraseMax <ms>   Microsecs for max blank between phrases, default: %i\n",arg_DelayPhraseMax);
      fprintf (stdout, "  --offsetx <num>         X offset in center mode, default: %f\n", arg_OffsetX);
      fprintf (stdout, "  --offsety <num>         Y offset in center mode, default: %f\n", arg_OffsetY);
      fprintf (stdout, "  --random                Show phrases in random order (default)\n");
      fprintf (stdout, "  --no-random             Show phrases in listed order\n");
      fprintf (stdout, "  --screensaver           Wait for an active screensaver\n");
      fprintf (stdout, "  --no-screensaver        Draw over active windows (default)\n");
      fprintf (stdout, "  --outline               Draw a contrasting outline around words (default)\n");
      fprintf (stdout, "  --no-outline            Draw words without an outline\n");
      fprintf (stdout, "  --center                Draw words in the center of the screen (default)\n");
      fprintf (stdout, "  --no-center             Draw words randomly around the screen\n");
      fprintf (stdout, "  --wordmode              Draw words one at the time\n");
      fprintf (stdout, "  --oneshot               Shows one phrase only than exits\n");
      exit (-1);
    }

  if (arg_DelayPhraseMax < arg_DelayPhraseMin)
    {
      arg_DelayPhraseMax = arg_DelayPhraseMin;
    }

  /* Get the phrases */
  text_Index = 0;
  text_Item = 0;
  text_Count = 0;
  memset (text_Used, 0, sizeof (text_Used));
  arg_Source = get_string_resource (dpy, XSUBLIM_ARG_FILE, "Filename");
  if (arg_Source != NULL)
    {
      FILE *file_Fs;
      struct stat file_Stat;

      file_Fs = fopen (arg_Source, "rb");
      if (file_Fs == NULL)
        {
          fprintf (stderr, "%s: Could not open '%s'\n", progname, arg_Source);
          exit (-1);
        }
      if (fstat (fileno (file_Fs), &file_Stat) != 0)
        {
          fprintf (stderr, "%s: Could not stat '%s'\n", progname, arg_Source);
          exit (-1);
        }
      arg_Text = calloc (1, file_Stat.st_size + 1);
      if (arg_Text != NULL)
        {
          if (fread (arg_Text, file_Stat.st_size, 1, file_Fs) != 1)
            {
              fprintf (stderr, "%s: Could not read '%s'\n", progname, arg_Source);
              exit (-1);
            }
        }
      fclose (file_Fs);
    }
  else
    {
      arg_Source = get_string_resource (dpy, XSUBLIM_ARG_PROGRAM, "Executable");
      if (arg_Source != NULL)
        {
          char *exe_Command = calloc (1, strlen (arg_Source) + 10);
          FILE *exe_Fs;

          if (exe_Command == NULL)
            {
              fprintf (stderr, "%s: Could not allocate space for '%s'\n", progname, arg_Source);
              exit (-1);
            }
          sprintf (exe_Command, "( %s ) 2>&1", arg_Source);

          exe_Fs = popen (exe_Command, "r");
          if (exe_Fs == NULL)
            {
              fprintf (stderr, "%s: Could not run '%s'\n", progname, arg_Source);
              exit (-1);
            }
          arg_Text = calloc (1, XSUBLIM_PROGRAM_SIZE);
          if (arg_Text != NULL)
            {
              if (fread (arg_Text, 1, XSUBLIM_PROGRAM_SIZE, exe_Fs) <= 0)
                {
                  fprintf (stderr, "%s: Could not read output of '%s'\n", progname, arg_Source);
                  exit (-1);
                }
              if (strstr (arg_Text, ": not found") ||
                  strstr (arg_Text, ": Not found") || strstr (arg_Text, ": command not found")
                  || strstr (arg_Text, ": Command not found"))
                {
                  fprintf (stderr, "%s: Could not find '%s'\n", progname, arg_Source);
                  exit (-1);
                }
            }
          fclose (exe_Fs);
        }
      else
        {
          arg_Text = get_string_resource (dpy, XSUBLIM_ARG_PHRASES, "Phrases");
          if (arg_Text != NULL)
            {
              arg_Text = strdup (arg_Text);
            }
        }
    }
  if (arg_Text != NULL)
    {
      while (((text_Phrase = strtok (arg_Text, "\n")) != NULL) && (text_Count < XSUBLIM_TEXT_COUNT))
        {
          arg_Text = NULL;
          text_List[text_Count] = text_Phrase;
          text_Count++;
        }
      text_List[text_Count] = NULL;
    }
  if (text_Count == 0)
    {
      fprintf (stderr, "%s: No text to display\n", progname);
      exit (-1);
    }

  /* Load the font */
  font_Font = XLoadQueryFont (dpy, get_string_resource (dpy, XSUBLIM_ARG_FONT, "Font"));
  font_Index = 0;
  while ((font_Font == NULL) && (font_List[font_Index] != NULL))
    {
      font_Font = XLoadQueryFont (dpy, font_List[font_Index]);
      font_Index++;
    }
  if (font_Font == NULL)
    {
      fprintf (stderr, "%s: Couldn't load a font\n", progname);
      exit (-1);
    }

  /* Create the GCs */
  XGetWindowAttributes (dpy, win_Root, &attr_Win);
  gc_ValFore.font = font_Font->fid;
  gc_ValFore.foreground = get_pixel_resource (dpy, attr_Win.colormap, "foreground", "Foreground");
  gc_ValFore.background = get_pixel_resource (dpy, attr_Win.colormap, "background", "Background");
  gc_ValFore.subwindow_mode = IncludeInferiors;
  gc_GcFore = XCreateGC (dpy, win_Root, (GCFont | GCForeground | GCBackground | GCSubwindowMode), &gc_ValFore);
  gc_ValBack.font = font_Font->fid;
  gc_ValBack.foreground = get_pixel_resource (dpy, attr_Win.colormap, "background", "Background");
  gc_ValBack.background = get_pixel_resource (dpy, attr_Win.colormap, "foreground", "Foreground");
  gc_ValBack.subwindow_mode = IncludeInferiors;
  gc_GcBack = XCreateGC (dpy, win_Root, (GCFont | GCForeground | GCBackground | GCSubwindowMode), &gc_ValBack);

  /* Loop ------------------------------------------------------------ */
  while (Xsublim_Sig_Last == -1)
    {
      /* Once-per-phrase stuff ----------------------------------- */

      /* If we're waiting for a screensaver... */
      if (arg_FlagScreensaver != FALSE)
        {
          /* Find the screensaver's window */
          win_Root = xsublim_Ss_GetWindow (dpy);
          if (win_Root == 0)
            {
              usleep (30000000);
              continue;
            }
        }

      /* Pick the next phrase */
      if (arg_FlagRandom != FALSE)
        {
          text_Item = random () % text_Count;
          text_Index = 0;
        }
      while (text_Used[text_Item] != FALSE)
        {
          text_Index++;
          text_Item++;
          if (text_Index == text_Count)
            {
              text_Index = 0;
              memset (text_Used, 0, sizeof (text_Used));
            }
          if (text_List[text_Item] == NULL)
            {
              text_Item = 0;
            }
        }
      text_Used[text_Item] = TRUE;
      strncpy (text_Text, text_List[text_Item], XSUBLIM_TEXT_LENGTH);
      text_Phrase = text_Text;

      /* Run through the phrase */
      if (arg_FlagWordmode == 1)
        text_Wordmode = " \t";
      else
        text_Wordmode = "\n";
      while (((text_Word = strtok (text_Phrase, text_Wordmode)) != NULL) && (Xsublim_Sig_Last == -1))
        {
          text_Phrase = NULL;

          /* Once-per-word stuff ----------------------------- */

          /* Find the text's position */
          XGetWindowAttributes (dpy, win_Root, &attr_Win);
          text_Length = strlen (text_Word);
          text_Width = XTextWidth (font_Font, text_Word, text_Length) + XSUBLIM_TEXT_OUTLINE * 2;
          text_Height = font_Font->ascent + font_Font->descent + 1 + XSUBLIM_TEXT_OUTLINE * 2;
          if (arg_FlagCenter == FALSE)
            {
              text_X = random () % (attr_Win.width - text_Width);
              text_Y = random () % (attr_Win.height - text_Height);
            }
          else
            {
              text_X = ((attr_Win.width*arg_OffsetX) / 2) - (text_Width / 2);
              text_Y = ((attr_Win.height*arg_OffsetY) / 2) - (text_Height / 2);
            }

          /* Find the image's position (and pad it out slightly,
             otherwise bits of letter get left behind -- are
             there boundry issues I don't know about?) */
          image_X = text_X - 16;
          image_Y = text_Y;
          image_Width = text_Width + 32;
          image_Height = text_Height;
          if (image_X < 0)
            {
              image_X = 0;
            }
          if (image_Y < 0)
            {
              image_Y = 0;
            }
          if (image_X + image_Width > attr_Win.width)
            {
              image_Width = attr_Win.width - image_X;
            }
          if (image_Y + image_Height > attr_Win.height)
            {
              image_Height = attr_Win.height - image_Y;
            }

          /* Influence people for our own ends --------------- */

          /* Grab the server -- we can't let anybody draw over
             us */
          XSync (dpy, FALSE);
          XGrabServer (dpy);
          XSync (dpy, FALSE);

          /* Set up an error handler that ignores BadMatches --
             since the screensaver can take its window away at
             any time, any call that uses it might choke */
          Xsublim_Sh_Status = 0;
          Xsublim_Sh_Handler = XSetErrorHandler (xsublim_Sh_Handler);

          /* Save the current background */
          image_Image = XGetImage (dpy, win_Root, image_X, image_Y, image_Width, image_Height, ~0L, ZPixmap);

          /* If we've successfully saved the background... */
          if (image_Image != NULL)
            {
              if (Xsublim_Sh_Status == 0)
                {
                  /* Draw the outline */
                  if (arg_FlagOutline != FALSE)
                    {
                      for (text_OutlineIndex = 0; text_Outline[text_OutlineIndex].outline_X != 0; text_OutlineIndex++)
                        {
                          /* Y'know, eight
                             character tabs and
                             descriptive variable
                             names become
                             annoying at some
                             point... */
                          XDrawString (dpy,
                                       win_Root, gc_GcBack,
                                       text_X +
                                       text_Outline
                                       [text_OutlineIndex].outline_X *
                                       XSUBLIM_TEXT_OUTLINE,
                                       text_Y + (font_Font->ascent) +
                                       text_Outline[text_OutlineIndex].outline_Y * XSUBLIM_TEXT_OUTLINE, text_Word,
                                       text_Length);
                        }
                    }

                  /* Draw the word */
                  XDrawString (dpy, win_Root, gc_GcFore, text_X, text_Y + (font_Font->ascent), text_Word, text_Length);
                }
              if (Xsublim_Sh_Status == 0)
                {
                  /* Wait a bit */
                  XSync (dpy, FALSE);
                  if (Xsublim_Sig_Last == -1)
                    {
                      usleep (arg_DelayShow);
                    }

                  /* Restore the background */
                  XPutImage (dpy, win_Root, gc_GcFore, image_Image, 0, 0, image_X, image_Y, image_Width, image_Height);
                }

              /* Free the image */
              XDestroyImage (image_Image);
            }

          /* Restore the error handler, ungrab the server */
          XSync (dpy, FALSE);
          XSetErrorHandler (Xsublim_Sh_Handler);
          XUngrabServer (dpy);
          XSync (dpy, FALSE);

          /* Pause between words */
          if (Xsublim_Sig_Last == -1)
            {
              usleep (arg_DelayWord);
            }
        }
      if (arg_FlagOneshot == TRUE) exit(0);
      
      /* Pause between phrases */
      if (Xsublim_Sig_Last == -1)
        {
          usleep (random () % (arg_DelayPhraseMax - arg_DelayPhraseMin + 1) + arg_DelayPhraseMin);
        }
    }

  /* Exit ------------------------------------------------------------ */
  for (sig_Number = 0; sig_Signal[sig_Number] != -1; sig_Number++)
    {
      signal (sig_Number, SIG_DFL);
    }
  kill (getpid (), Xsublim_Sig_Last);

  return 0;
}
